/* 
 *   Creation Date: <2002/01/13 20:45:37 samuel>
 *   Time-stamp: <2003/01/27 23:53:43 samuel>
 *   
 *	<init.c>
 *	
 *	Kernel module initialization
 *   
 *   Copyright (C) 2002, 2003 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#include "archinclude.h"
#include "alloc.h"
#include "kernel_vars.h"
#include "misc.h"
#include "multiplexer.h"
#include "mmu.h"
#include "asmfuncs.h"
#include "performance.h"

/* Globals */
session_table_t		*g_sesstab = NULL;
int 			g_num_sessions=0;


/************************************************************************/
/*	Performance Info						*/
/************************************************************************/

//#define HOTSPOT

#ifdef PERFORMANCE_INFO

void
clear_performance_info( kernel_vars_t *kv )
{
	perf_info_t *p = g_perf_info_table;
	int i;
#ifdef HOTSPOT_INFO
	extern long r__hotspot_table[], r__hotspot_table_end[];
	long *pp = (long*)reloc_ptr( r__hotspot_table );
	int n = ((ulong)r__hotspot_table_end - (ulong)r__hotspot_table) / sizeof(ulong[2]);

	for( i=0; i<n && pp[i] ; i+=2 )
		pp[i]=pp[i+1]=0;
#endif	
	for( ; p->name ; p++ )
		*p->ctrptr = 0;
	for( i=0; i<NUM_ASM_BUMP_CNTRS; i++ )
		kv->asm_bump_cntr[i] = 0;
	kv->num_acntrs = 0;
}

#ifdef HOTSPOT_INFO
static void
find_hotspot( void )
{
	extern long r__hotspot_table[], r__hotspot_table_end[];
	long *p = (long*)reloc_ptr( r__hotspot_table );
	int n = ((ulong)r__hotspot_table_end - (ulong)r__hotspot_table) / sizeof(ulong[2]);
	int i,j, maxind=0;
	ulong total=0;
	
	printk("\n");
	for( j=0; j<20; j++ ) {
		for( i=0; i<n && p[i] ; i+=2 ) {
			if( p[i+1] > p[maxind+1] ) 
				maxind = i;
		}
		printk("nip: %08lX, count %08lX (%ld)\n",
		       p[maxind], p[maxind+1], p[maxind+1] );
		total += p[maxind+1];
		p[maxind+1] = 0;
	}
	printk("total = %08lX (%08ld)\n", total, total );
}
#endif


int
get_performance_info( kernel_vars_t *kv, int ind, char *buf, int size, int *retval )
{
	perf_info_t *p = g_perf_info_table;
	int i, j, len, ctr;
	char *name;

#ifdef HOTSPOT_INFO
	if( !ind )
		find_hotspot();
#endif	
	for( i=0; p->name && i!=ind; i++, p++ )
		;
	if( !p->name ) {
		extern char *__start_bumptable[], *__end_bumptable[];
		char **s;
		for( j=0, s=__start_bumptable ; s < __end_bumptable && i!=ind ; s++, i++, j++ )
			;
		if( s >= __end_bumptable )
			return 1;
		name = *s;
		ctr = kv->asm_bump_cntr[j];
	} else {
		name = p->name;
		ctr = *p->ctrptr;
	}
	
	len = strlen(name)+1;
	if( len > size )
		len = size;
	if( copy_to_user_mol( buf, name, len ) || copy_int_to_user( retval, ctr ))
		return 1;
	return 0;
}

#else /* PERFORMANCE_INFO */

void
clear_performance_info( kernel_vars_t *kv ) 
{
}

int
get_performance_info( kernel_vars_t *kv, int ind, char *buf, int size, int *retval ) 
{ 
	return 1;
}

#endif /* PERFORMANCE_INFO */


/************************************************************************/
/*	init/cleanup kernel module					*/
/************************************************************************/

int
common_init_module( void )
{
	if( !(g_sesstab = kmalloc_cont_mol( sizeof(session_table_t))) )
		return 1;
	memset( g_sesstab, 0, sizeof(session_table_t) );

	if( relocate_code() ) {
		kfree_cont_mol( g_sesstab );
		return 1;
	}

	printk( "MOL %d.%d.%d kernel module loaded\n", 
		MAJOR_VERSION, MINOR_VERSION, PATCH_LEVEL );
	return 0;
}

void
common_cleanup_module( void )
{
	int i;

	for(i=0; i<MAX_NUM_SESSIONS; i++ )
		destroy_session( i );

	remove_hooks();

	relocation_cleanup();
	kfree_cont_mol(g_sesstab);
}


/************************************************************************/
/*	initialize / destroy session					*/
/************************************************************************/

int 
initialize_session( int index, ulong session_magic )
{
	kernel_vars_t *kv;

	/* printk("initialize_session\n" ); */

	if( index < 0 || index > MAX_NUM_SESSIONS )
		return errSessionOutOfBounds;
	if( g_sesstab->kvars[index] )
		return errSessionInUse;

	if( !g_num_sessions && write_hooks() )
		return errGeneralError;

	if( !(kv = alloc_kvar_pages()) )
		return errGeneralError;

	memset( kv, 0, NUM_KVARS_PAGES * 0x1000 );
	kv->kvars_virt = kv;

	clear_performance_info( kv );
#ifdef __linux__
	if( init_mmu( kv ) ) {
		free_kvar_pages( kv );
		return errGeneralError;
	}
#endif
	msr_altered( kv );
	inc_mod_count();

	g_num_sessions++;

	g_sesstab->kvars_ph[index] = tophys_mol(kv);
	g_sesstab->kvars[index] = kv;
	g_sesstab->entry_magic[index] = session_magic;
	return 0;
}

void
destroy_session( int index )
{
	kernel_vars_t *kv;
	/* printk("Destroy session\n"); */

	if( index < 0 || index > MAX_NUM_SESSIONS || !(kv=g_sesstab->kvars[index]) )
		return;

	g_sesstab->kvars[index] = NULL;
	g_sesstab->kvars_ph[index] = 0;
	g_sesstab->entry_magic[index] = 0;

	/* Decrease before freeing anything (simplifies deallocation of shared resources) */
	g_num_sessions--;
#ifdef __linux__
	cleanup_mmu( kv );
#endif
	if( kv->emuaccel_page )
		free_page_mol( kv->emuaccel_page );

	memset( kv, 0, NUM_KVARS_PAGES * 0x1000 );

	free_kvar_pages( kv );

	if( !g_num_sessions )
		remove_hooks();

	dec_mod_count();
}
